/*
** Module   : ExternalProgramWrapper.java
** Abstract : container that implements persistent-procedure related APIs.
**
** Copyright (c) 2012-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 002 CA  20121220 Created initial version.
** 003 CS  20130211 Added conversion support for app server. 
** 004 CA  20130221 getType getter for TYPE attribute was renamed to getResourceType. 
**                  Added ADM-DATA and UNIQUE-ID support.
** 005 CS  20130221 Added conversion support for CURRENT-WINDOW attribute.
** 006 CA  20130223 Added stubs for SET-CALLBACK-PROCEDURE method.
** 007 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 008 CA  20130529 Added implementation for REMOTE, SERVER-HANDLE and PROXY attributes. 
**                  Implemented the resource delete here.
** 009 CA  20130813 Added support for the ASYNC-REQUEST-COUNT attribute.
** 010 CA  20130919 Resource cleanup must be done after it was deleted.
** 011 CA  20130925 Added support for PUBLISHED-EVENTS attribute.
** 012 CA  20130927 Resource type is determined from LegacyResource constant; the annotation is 
**                  ignored, as it doesn't extend HandleResource. 
** 013 EVL 20131014 Adding annotation for legacy attributes to be ignored for the class.
** 014 SVL 20131203 Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 015 CA  20131013 Added support for INSTANTIATING-PROCEDURE attribute.
** 016 CA  20131208 Emit SEARCH-TARGET and SEARCH-SELF literals as integer.  Do note delete the 
**                  procedure if is active on the call stack.
** 017 HC  20131222 Implemented ADM-DATA and UNIQUE-ID attribute.
**                  INSTANTIATING-PROCEDURE is now managed by ProcedureManager.
** 018 VIG 20140108 Implemented CURRENT-WINDOW attribute.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.ui.LogicalTerminal;
import com.goldencode.p2j.ui.WindowWidget;
import com.goldencode.p2j.util.ProcedureManager.SearchMode;

/**
 * This container adds support for all persistent-procedure related attributes
 * and methods, such as:
 * <ul>
 *   <li>INTERNAL-ENTRIES</li>
 *   <li>NAME</li>
 *   <li>FILE-NAME</li>
 *   <li>PERSISTENT</li>
 *   <li>TYPE</li>
 *   <li>NEXT-SIBLING</li>
 *   <li>PREV-SIBLING</li>
 *   <li>PRIVATE-DATA</li>
 *   <li>GET-SIGNATURE</li>
 *   <li>ADD-SUPER-PROCEDURE</li>
 *   <li>REMOVE-SUPER-PROCEDURE</li>
 *   <li>SUPER-PROCEDURES</li>
 *   <li>SERVER</li>
 *   <li>ASYNC-REQUEST-COUNT</li>
 *   <li>INSTANTIATING-PROCEDURE</li>
 *   <li>CURRENT-WINDOW</li>
 * </ul>
 */
public class ExternalProgramWrapper
extends TransparentWrapper
implements PersistentProcedure
{
   /**
    * Constructs a new wrapper for the given external procedure instance.
    *  
    * @param    referent
    *           The external procedure instance.
    */
   public ExternalProgramWrapper(Object referent)
   {
      super(referent);
   }

   /**
    * Determine if this instance is valid. This means that its 
    * {@link TransparentWrapper#referent} field must be set and valid. More,
    * this must be a known procedure instance, registered with 
    * {@link ProcedureManager}.
    * 
    * @return   <code>true</code> if this instance is valid, 
    *           <code>false</code> otherways.
    */
   public boolean valid()
   {
      if (!super.valid())
      {
         return false;
      }
      
      return ProcedureManager.hasReferent(referent);
   }
   
   /**
    * API needed to implement read-only attribute assignment (a 4GL 
    * "feature").
    * 
    * @param    attribute
    *           The attribute's name.
    */
   public void readOnlyError(String attribute)
   {
      handle.readOnlyError(new handle(this), attribute);
   }
   
   /**
    * Get the <code>INTERNAL-ENTRIES</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public character internalEntries()
   {
      String pname = getFileName().toStringMessage();
      boolean prv = this.referent == ProcedureManager.thisProcedure().get();

      return new character(SourceNameMapper.getInternalEntries(pname, prv));
   }

   /**
    * Get the <code>NAME</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public character name()
   {
      return getFileName();
   }

   /**
    * Calls {@link handle#readOnlyError}, as this attribute is read-only for procedure handles. 
    * 
    * @param    name
    *           Not used.
    */
   @Override
   @LegacyAttribute(name = "NAME", setter = true, ignore = true)
   public void name(String name)
   {
      handle.readOnlyError(new handle(this), "name");
   }

   /**
    * Calls {@link handle#readOnlyError}, as this attribute is read-only for procedure handles. 
    * 
    * @param    name
    *           Not used.
    */
   @Override
   @LegacyAttribute(name = "NAME", setter = true, ignore = true)
   public void name(character name)
   {
      name((String) null);
   }

   /**
    * Get the <code>FILE-NAME</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public character getFileName()
   {
      return ProcedureManager.getFileName(referent);
   }

   /**
    * Get the <code>persistent</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public logical isPersistent()
   {
      return ProcedureManager.isPersistent(referent);
   }
   
   /**
    * Get the <code>TYPE</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   Always returns <code>"PROCEDURE"</code>.
    */
   public character getResourceType()
   {
      return new character(LegacyResource.PROCEDURE);
   }

   /**
    * Get the <code>NEXT-SIBLING</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public handle getNextSibling()
   {
      return ProcedureManager.getNextSibling(referent);
   }

   /**
    * Get the <code>PREV-SIBLING</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public handle getPrevSibling()
   {
      return ProcedureManager.getPrevSibling(referent);
   }

   /**
    * Get the <code>PRIVATE-DATA</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public character getPrivateData()
   {
      return ProcedureManager.getPrivateData(referent);
   }

   /**
    * Set the <code>PRIVATE-DATA</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @param    data
    *           The new private-data value. 
    */
   public void setPrivateData(character data)
   {
      ProcedureManager.setPrivateData(referent, data);
   }

   /**
    * Set the <code>PRIVATE-DATA</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @param    data
    *           The new private-data value. 
    */
   public void setPrivateData(String data)
   {
      setPrivateData(new character(data));
   }

   /**
    * Execute the <code>GET-SIGNATURE</code> method of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @param    internalEntry
    *           The legacy name of the internal-entry for which the signature
    *           is needed.
    * 
    * @return  The internal entry's signature.
    */
   public character getSignature(String internalEntry)
   {
      return getSignature(new character(internalEntry));
   }

   /**
    * Execute the <code>GET-SIGNATURE</code> method of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @param    internalEntry
    *           The legacy name of the internal-entry for which the signature
    *           is needed. 
    * 
    * @return  The internal entry's signature.
    */
   public character getSignature(character internalEntry)
   {
      String ieName = internalEntry.toStringMessage();
      character fileName = getFileName();
      String pname = fileName.toStringMessage();

      if ("".equals(ieName))
      {
         return new character(SourceNameMapper.getSignature(pname, ""));
      }

      // is safe to call both, because: if there are both func and proc with
      // the same name, then the signature is for the func, not the proc
      // (thus if the func is private, then is not seen).
      // else, if no func with this name, proc is cheched
      if (SourceNameMapper.isPrivate(pname, ieName, true) ||
          SourceNameMapper.isPrivate(pname, ieName, false))
      {
         // am I calling this from THIS-PROCEDURE? if yes, then I have access
         handle thisp = ProcedureManager.thisProcedure();
         character thisName = handle.unwrapProcedure(thisp)
                                    .getFileName();
         if (!CompareOps.equals(fileName, thisName))
         {
            // return blank
            return new character("");
         }
      }

      return new character(SourceNameMapper.getSignature(pname, ieName));
   }

   /**
    * Get the <code>SUPER-PROCEDURES</code> attribute of this 
    * {@link TransparentWrapper#referent external procedure instance.}
    * 
    * @return   See above.
    */
   public character superProcedures()
   {
      return ProcedureManager.superProcedures(referent);
   }

   /**
    * Add the given handle as a super-procedure for the current handle.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public logical addSuperProcedure(handle h)
   {
      return ProcedureManager.addSuperProcedure(referent, h);
   }

   /**
    * Add the given handle as a super-procedure for the current handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public logical addSuperProcedure(handle h, SearchMode mode)
   {
      return ProcedureManager.addSuperProcedure(referent, h, mode);
   }

   /**
    * Add the given handle as a super-procedure for the current handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public logical addSuperProcedure(handle h, int mode)
   {
      return ProcedureManager.addSuperProcedure(referent, h, mode);
   }

   /**
    * Add the given handle as a super-procedure for the current handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public logical addSuperProcedure(handle h, integer mode)
   {
      if (mode.isUnknown())
      {
         return ProcedureManager.addSuperProcedure(referent, h, (SearchMode) null);
      }
      else
      {
         return ProcedureManager.addSuperProcedure(referent, h, mode.intValue());
      }
   }

   /**
    * Remove the given handle from this handle's list of super-procedures.
    *  
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public logical removeSuperProcedure(handle h)
   {
      return ProcedureManager.removeSuperProcedure(referent, h);
   }

   /**
    * Get CURRENT-WINDOW attribute for given procedure handle.
    *          
    * @return  Reference to current window widget.
    */
   public handle currentWindow()
   {
      handle phandle = new handle(this);
      // TODO: handle validation
      return LogicalTerminal.currentWindow(phandle);
   }

   /**
    * Set CURRENT-WINDOW attribute for this procedure handle.
    * 
    * @param    h
    *           A handle for the window.
    */
   public void setCurrentWindow(handle h)
   {
      handle phandle = new handle(this);
      LogicalTerminal.setProcCurrentWindow(phandle, h);
   }

   /**
    * Set CURRENT-WINDOW attribute for this procedure handle.
    * 
    * @param    win
    *           Window reference.
    */
   public void setCurrentWindow(WindowWidget win)
   {
      setCurrentWindow(win.asWidgetHandle());
   }
   
   /**
    * Indicates if the procedure is ran as the remote top-level AppServer.
    * 
    * @return   <code>true</code> if the procedure is ran remote , <code>false</code> otherwise.
    */
   public  logical isRemote()
   {
      return new logical(SessionUtils._isRemote() && 
                         (ProcedureManager.isTopLevel(get()) ||
                          AppServerManager.hasProcedure(this)));
   }
   
   /**
    * Returns a handle to the app server in which the procedure is ran.
    * 
    * @return   Always unknown.
    */
   public handle getServerHandle()
   {
      return new handle();
   }

   /**
    * Indicates if the procedure is a proxy procedure.
    * 
    * @return   always <code>false</code>
    */
   public logical isProxy()
   {
      return new logical(false);
   }
   
   /**
    * Returns a handle to the current opened transaction.
    * 
    * @return   See above
    */
   public handle getTransaction()
   {
      // TODO implement actual logic
      return null;
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return ProcedureManager.getADMData(referent);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      ProcedureManager.setADMData(referent, value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      ProcedureManager.setADMData(referent, value.getValue());
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      return ProcedureManager.getUniqueID(referent);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, String proc)
   {
      return setCallbackProcedure(new character(event), new character(proc),
                                  ProcedureManager.thisProcedure());
   }
   
   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, String proc)
   {
      return setCallbackProcedure(event, new character(proc), ProcedureManager.thisProcedure());
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, character proc)
   {
      return setCallbackProcedure(new character(event), proc, ProcedureManager.thisProcedure());
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event, which will be executed using
    *           the current THIS-PROCEDURE reference. To remove a callback mapping, set the this
    *           parameter to the empty string.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, character proc)
   {
      return setCallbackProcedure(event, proc, ProcedureManager.thisProcedure());
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, String proc, handle context)
   {
      return setCallbackProcedure(new character(event), new character(proc), context);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, String proc, handle context)
   {
      return setCallbackProcedure(event, new character(proc), context);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(String event, character proc, handle context)
   {
      return setCallbackProcedure(new character(event), proc, context);
   }

   /**
    * Associate the internal procedure with the given web-services event.
    * <p>
    * This method returns false if the event name is not valid or if the procedure's context is
    * not a valid handle.
    * 
    * @param    event
    *           The event name, must evaluate to one of the "REQUEST-HEADER" or the 
    *           "RESPONSE-HEADER" strings. 
    * @param    proc
    *           The internal procedure to be executed for this event. To remove a callback 
    *           mapping, set the this parameter to the empty string.
    * @param    context
    *           The procedure handle where the internal procedure will be searched and executed.
    *           
    * @return   <code>true</code> if the event was registered.
    */
   @Override
   public logical setCallbackProcedure(character event, character proc, handle context)
   {
      // TODO: implement this
      return null;
   }

   /**
    * Perform actual delete of this procedure, by calling {@link ProcedureManager#delete(Object)}.
    */
   public void delete()
   {
      if (ProcedureManager.delete(this.get()))
      {
         // cleanup after it
         handle.removeResource(this);
      }
      
      // if this is an appserver REMOTE procedure handle, then the requester side is not informed
      // about the handle deletion; only when deleting it from the requester side, the remote side
      // will be informed
   }

   /**
    * Returns the number of asynchronous requests.
    * 
    * @return   For non-proxy procedures, always 0.
    */
   @Override
   public integer getAsyncRequestCount()
   {
      return new integer(0);
   }
   
   /**
    * Get the named events published by this procedure.  This is a comma separated list of
    * named events collected from the PUBLISH statements, where the named event is a hard-coded
    * string, not a variable.
    * 
    * @return    The named events published by this procedure, or the empty string if none found.
    */
   @Override
   public character getPublishedEvents()
   {
      String pname = ProcedureManager.getFileName(referent).toStringMessage();
      return new character(SourceNameMapper.getPublishedEvents(pname));
   }

   /**
    * Getter for the INSTANTIATING-PROCEDURE attribute.
    * <p>
    * The returned handle will refer a valid external procedure (persistent or not) or will be
    * unknown, when this is a entry point procedure or in case its instantiating procedure was 
    * persistent and has been deleted.
    * 
    * @return   See above.
    */
   @Override
   public handle instantiatingProcedure()
   {
      return ProcedureManager.resolveInstantiatingProcedure(ProcedureManager.getInstantiatingProcedure(referent));
   }
}